\     Turtle-Logo in Forth, Copyright (c) 2016-2018, Assad Ebrahim, Mathematical Science & Technologies
\          FREE download from: http://www.mathscitech.org/articles/turtle-logo-forth
\          Contact: assad.ebrahim@alum.swarthmore.edu for comments, bug-reports, feature-requests

\     GNU License - Users are permitted free re-use provided:                          
\           1/ Source code for all modified versions are also made freely available under the GNU License            
\           2/ Attributions are preserved (lines 1-6 and Revision History)                          

\     Description: This is a Turtle logo program for Windows operated entirely with a few keys.  
\                  It has several advanced features, including macro recording and replay.
\                  It is written in ~750 lines of Forth code for the F-PC Forth for DOS.
\                  To run it requires a DOS emulator, either vDOSplus (recommended) or DOSbox (both freely available)
\                  A complete run-time package with Windows installer at the website above.  

\     Turnkey Startup: Run vDOSplus.exe which in turn calls "turtle_logo.bat".
\                      Further instructions are provided in the _readme.txt

\     Developers Corner: To modify the program, you'll need to edit turtle.fs and run it in F-PC (Forth FPC) using command:
\                                   fload turtle.fs, then: turtle-go
\                        To package a modified release, recompile F-PC with FMETA to create F-PC with turtle.fs embedded in it (turnkey application)
\                        F-PC can be downloaded from: ftp://ftp.taygeta.com/pub/Forth/Compilers/native/dos/FPC/fpc36.zip
\                        Instructions included in docs: Fpc\doc\ting_1989_FPC_UsersManual\Chapter9_Rebuilding_FPC.txt


\ Revision History
: rev# ." 1.5.4021" ;
: welcome ( -- )
       ." Welcome to Turtle Logo! (" rev# ." , (c) 2016-2018, Assad Ebrahim)"
    cr
    cr ." QUICK START:"
    cr ."   (1) To MOVE Turtle, use the arrow keys and backspace..."
    cr ."     FWD moves one step along current heading, and paints."
    cr ."     TURN LEFT/RIGHT changes Turtle's heading 45 deg counter-/clock- wise"
    cr ."     BACKSPACE backs up and erases."                   \ Note: movement wraps around the edges of the screen to the other side (Exercise)
    cr ."   (2) To CHANGE COLOR... "
    cr ."     C / X cycles pen color (fwd/back) through 16 choices."
    cr ."     ' toggles pen up (to move without painting) or down (to draw again)"
    cr ."     H sends Turtle to home position at bottom left of screen."
    cr ."     ; toggles Turtle's visibility (hide or show)."                            \ instruction not on main screen
    cr ."   (3) To TEACH Turtle some tricks, you can have him learn and repeat some steps..."  \ RECORD/REPLAY macros..."
    cr ."     T to begin/end TEACHING Turtle a particular SKILL. (Turtle can learn 9 skills.)"       \ skill=diary.  can record nested macros, but not recursive macros
    cr ."     N cycles through skills 1-9.  Pressing 1-9 gets Turtle to replay that skill."     \ Note, it will be EXACTLY what you told him. \ Later add LOOP
    cr ."   (4) Other actions..."
    cr ."     ESC clears screen -- once you confirm."
    cr ."     ? shows this HELP screen again (your drawing is safe!)"
    cr ."     Q quits the program (after confirming)"
    cr ." Happy Turtle-ing!  
    cr
    cr ." For ideas & challenge problems, see http://www.mathscitech.org/articles/turtle-logo-forth"
    cr ;

\ ============ Control keys
\ ASCII codes for user keyboard commands
\ Child friendly: Upper and lower case letters are added so program still works even if Caps Locks is pressed (note, numbers are not affected by Caps Lock)
\ / is treated as identical to ? for same reason (help)
\ Remark: you will get insite when looking at ASCII codes in hex (binary) not decimal
$C8          constant    fwd-key      \ 200d
$CD          constant    right-key    \ 205d
right-key    constant    cw-key
$CB          constant    left-key     \ 203d
left-key     constant    ccw-key
$8           constant    backspace
backspace    constant    erase-key
$27          constant    tick          \ 39d
tick         constant    penupdown-key      
'c'          constant    color-fwd-key      \ change color
'C'          constant    color-fwd-key2     \ change color
'x'          constant    color-back-key     \ change color
'X'          constant    color-back-key2    \ change color
'n'          constant    diary-key          \ RENAMED NEXT
'N'          constant    diary-key2         \ RENAMED NEXT
't'          constant    record-key         \ RENAMED TEACH
'T'          constant    record-key2        \ RENAMED TEACH
'1'          constant    play-macro-1-key
'2'          constant    play-2-key
'3'          constant    play-3-key
'4'          constant    play-4-key
'5'          constant    play-5-key
'6'          constant    play-6-key
'7'          constant    play-7-key
'8'          constant    play-8-key
'9'          constant    play-9-key
';'          constant    show-hide-key      \ turtle hide/view 
'h'          constant    home-key 
'H'          constant    home-key2
$1B          constant    esc                 \ 27d
esc          constant    clear-screen-key
'?'          constant    help-key
'/'          constant    help-key2           \ Usability: in case a child does not know how to use shift key
'q'          constant    quit-key
'Q'          constant    quit-key2


\ === Interactivity / Scaffolding
: pause-key ( -- ) ." PRESS ANY KEY TO CONTINUE... " key drop ;
: pause-quit ( -- ) key 'q' over = swap 'Q' = or if abort then ;
: pq ( -- ) pause-quit ;   \ abbreviation - useful for single stepping
: prompt-clear ( -- ) ." Clear screen? (Press ESC to confirm, or any key to continue drawing.)" ; 
: clear? ( -- t/f )   \ y or ESC are confirmation
    prompt-clear key 'y' over = ( asc t/f ) over 'Y' = or ( asc t/f) swap esc = or
    if true else false then ;
: prompt-quit ( -- ) ." ...sure? (y or q to really quit, else any key to return.)"  ;
: quit? ( -- t/f )   \ q or y are confirmation
    key 
    'y' over = ( asc t/f ) over 'Y' = or ( asc t/f) over 'q' = or swap 'Q' = or
    if true else false then ;
: arrow ( -- )  '-' emit '>' emit ;
: show-key ( asc -- ) dup . arrow space emit ;

\ === Misc
: asc2digit ( asc -- digit )  '0' - ; 
: digit2asc ( digit -- asc )  '0' + ;

\ === Memory
: bytes     ( n -- n )  ;                                 \ 1x, for documentation
: cells     ( num -- num*wsz )  2 *  ;                    \ 16-bit word length on 8086 chip
: c@+       ( adr -- adr+ val )   dup 1 + swap c@ ;       \ fetch char and advance
: @+        ( adr -- adr+ val )   dup 1 cells + swap @ ;  \ fetch word and advance
: c!+       ( val adr -- adr+ )   dup 1 + -rot c! ;       \ store char and advance
: !+        ( val adr -- adr+ )   dup 1 cells + -rot ! ;  \ store word and advance

\ === Strings
$22   constant    dquote    \ " double quote
: s" ( | <wordstream"> -- c-addr nchar  ) dquote word c@+ pad swap 2dup >r >r cmove r> r> ;  \ create string from wordstream input
: cmove+ ( c-addr ptr n -- ptr' )  >r tuck r@ cmove r> + ;  \ copy byte sequence/string (c-addr, n bytes) to ptr, returning new position
: cmove-fwBL+ ( adr1 adr2 n fwlen -- adr' )                 \ copy adr1 to adr2 (n bytes) padding rest with BL to create fixed width field
    >r >r tuck r@ cmove r>   \ copy str to array
    2dup + ( array n array' ) 
    r@ rot - BL fill                          \ pad rest of fixed width field with blanks
    r> + ;                                    \ advance array    
: attach-string-fw ( d-adr fwlen str n -- adr' fwlen ) >r rot r> 3 pick cmove-fwBL+ swap ;   \ attach a fixed with string to an array

: get-str ( index array strlen -- ptr len ) rot over * rot + swap ;  \ Returns string ptr/len at index when working with an array of fixed length strings

: create-zstring    \ creates counted string (zstring) with character count stored in first byte   \ USE TO REPLACE the fiddly COLOR-NAMES design
\ Usage: : tmp " Hello World!" ;
\        tmp create-str str
\        str type
    create  ( addr n  <name> -- ptr )
    here >R 
    dup 1 + bytes allot   ( adr n )  \ counted string
    dup R> c!+ ( ptr n str' )
    swap cmove
does> ( -- strptr nchar ) 
    c@+ ;


\ === Color encoding (Text mode): hi nybble holds background color, low nybble holds foreground color
: shl4      ( val -- val )        $10 * ;                 \ raise nybble
: shr4      ( val -- val )        $10 / ;                 \ lower nybble
: fgbg2col ( fgcolor bgcolor -- colorcode )  shl4 + ;
: col2fgbg ( colorcode -- fgcolor bgcolor )  dup $0F and  ( mask out fgcolor )  swap shr4  ( downshift out bgcolor ) ;

\ =========  Colors
16               constant   num-colors
num-colors       constant   penup-col-index
num-colors 1 +   constant   num-color-str    \ 1 more for penup string
 0      constant  black
 1      constant  navy
 2      constant  green
 3      constant  teal  
 4      constant  red   
 5      constant  purple 
 6      constant  brown 
 7      constant  silver  
 8      constant  slate 
 9      constant  blue  
 10     constant  lime  
 11     constant  aqua  
 12     constant  orange
 13     constant  pink  
 14     constant  yellow
 15     constant  white 
 silver constant  grey     \ alternate spellings
 silver constant  gray
 slate  constant  darkgrey
 slate  constant  darkgray

\ ===  Color Name Strings
6        constant    strlen-col

: attach-colors ( c-addr L n | <wordstream> -- ) \ creates fixed width array of color names for use with menus, from input stream
    0 do BL word c@+ attach-string-fw loop 2drop ;

: create-color-names                         \ data structure holding color names as fixed width strings, retrieved by index #
create  ( L n | <name> <wordstream> --  )    \ color names: n (num-color-str) color strings of fixed width L (strlen-col)
    here -rot                                \ store object ptr
    2dup * 2 + bytes allot ( addr L n )      \ allocate storage for strings+metadata
    rot >r 2dup swap r> c!+ c!+              \ encode metadata L n at start of structure
    -rot ( c-addr L n )  attach-colors
does> ( index -- c-addr nchar )              \ get fixed width string at index i
    c@+ -rot 1 + -rot                        \ extract fw-size and get past metadata
    ( c-adr L index )  over * rot + swap ;   \ calculate str ptr 

strlen-col num-color-str create-color-names color-names Black Navy Green Teal Red Purple Brown Silver Slate Blue Lime Aqua Orange Pink Yellow White PEN-UP
\ Color comments: navy=dark blue, teal=aqua, red=maroon, silver=grey, slate=darkgrey, blue=lt blue, lime=lt green, aqua=lt aqua/cyan, pink=magenta/fuschia

: test-color-names ( -- ) color-names c@+ . c@ + . strlen-col num-color-str * type ;  \ test with blank does> 
: test-color-names2 ( -- ) 5 color-names type 8 color-names type ;


create screen-color 1 bytes allot                   \ VARIABLE - holds canvas color codes 0..15
create cursor-color 1 bytes allot                   \ VARIABLE - holds cursor color (bg color)

: set-screen-color ( color -- ) screen-color c! ;
: get-screen-color ( -- color ) screen-color c@ ;
: set-cursor-color ( bgcolor -- ) cursor-color c! ;
: get-cursor-color ( -- bgcolor ) cursor-color c@ ;

white  set-screen-color    \ default
silver set-cursor-color   \ default

\ === Pen/Cursor color mapping (foreground/background).
\ The design chooses an optimal cursor color (bgcolor) for each pen-color (fgcolor) to maximize their contrast ratio (see Table in documentation)
\ Rationale: We want color of pen (fgcolor) to contrast sufficiently with background to discern.  
\ Method: We use theory of color contrasts to choose the optimal pairing (equivalent to designer-provided) using color contrast checker tool (contrast ratio)
\ References:   http://www.worqx.com/color/color_contrast.htm
\               http://webaim.org/resources/contrastchecker/
\               http://juicystudio.com/services/luminositycontrastratio.php#specify
\ (ALTERNATIVE: could have it be paint color, then would need fg char to be high contrast or flashing)

\ fgbg lookup table is indexed by fgcolor and holds bgcolor (cursor) to give the optimal combination 
create  fgbg-lookup  num-colors  bytes  allot     
: fgbg!  ( fg bg -- )  swap  fgbg-lookup  +  c! ;      \ store bg color into table indexed by fg
: set-fgbg-palette ( -- )                              \ set optimized color palette
    \ fg    bg
    black    silver  fgbg!
    navy     aqua    fgbg!
    green    lime    fgbg!   \ not great, but ok
    teal     aqua    fgbg!
    red      yellow  fgbg!
    purple   white   fgbg!
    brown    silver  fgbg!
    silver   black   fgbg!
    slate    silver  fgbg!
    blue     aqua    fgbg!
    lime     slate   fgbg!
    aqua     navy    fgbg!
    orange   yellow  fgbg!
    pink     black   fgbg!
    yellow   brown   fgbg!
    white    black   fgbg! ;
\ penup is handled with specific penup color instructions

: get-optim-bg-color ( fg -- bg ) fgbg-lookup + c@ ;
: set-optim-bg-color ( pen-color-fg -- ) get-optim-bg-color set-cursor-color ;

create pen-color 1 bytes allot  \ VARIABLE -- pencolor is foreground color, 0..15
: get-pen-color ( -- c )  pen-color c@ ;
: set-pen-color ( c -- )  dup   pen-color c! set-optim-bg-color ;      \ automatically change background so turtle can be seen better
: set-penup-color ( -- )  white pen-color c! lime set-cursor-color ;   \ giving penup a distinct appearance, stored in last slot of fgbg table
: turtle-color ( asc -- ) color-fwd-key = if 1 else -1 then get-pen-color  +  16  mod  set-pen-color ;  \ +1/-1 cycle fwd/bk to next color of 16

create penup-state  1 cells allot  \ VARIABLE -- BOOLEAN true if pen is up else false
: set-penup-state   ( -- ) true penup-state c! ; 
: clear-penup-state ( -- ) false penup-state c! ;
: get-penup-state   ( -- ) penup-state c@ ;

create prev-pen-color 1 bytes allot  \ VARIABLE
: save-prev-pen-color    ( -- )   get-pen-color prev-pen-color c! ;
: restore-prev-pen-color ( -- c ) prev-pen-color c@ set-pen-color ;

: do-penup ( -- )  set-penup-state  save-prev-pen-color set-penup-color ; \ pen-up mode has custom appearance
: do-pendown ( -- )  clear-penup-state  restore-prev-pen-color  ;


\ States describing status of Recording (on or off or full etc.)
create rec-state 1 bytes allot
0 constant rec-state-ready
1 constant rec-state-recording
2 constant rec-state-full
3 constant rec-state-reminder  \ still recording, i.e. a flash
: set-rec-state ( r -- ) rec-state c! ;
: get-rec-state ( -- r ) rec-state c@ ;

\ Record strings
4       constant   nrec-msg                                \ number of different recording messages
12      constant   strlen-rec                              \ msg length
create  rec-msg-names    nrec-msg  strlen-rec  *  bytes  allot   \ VARIABLE recording string
: clear-rec-msg-names ( -- ) rec-msg-names nrec-msg strlen-rec * BL fill ;
clear-rec-msg-names
: show-rec-msg-names ( -- ) rec-msg-names nrec-msg strlen-rec * type cr pq ;
: attach-rec-msgs ( -- )  
    rec-msg-names strlen-rec
    " TeachSkill 1"     attach-string-fw   \ #1 is default diary/skill, update-diary-string keeps these current and updates menu dynamically
    " LEARNING...1"     attach-string-fw
    "  Full!"           attach-string-fw
    " LEARNING...1"     attach-string-fw 
    2drop ;
: get-rec-str ( -- ptr n ) get-rec-state rec-msg-names strlen-rec get-str ; 
11 constant diary-num-pos   \ diary num position in string
: update-diary-string ( digit -- ) 1 + digit2asc rec-msg-names diary-num-pos + ( val ptr ) 2dup c!  \ first string
        strlen-rec + 2dup c!             \ second string
        strlen-rec + strlen-rec + c! ;   \ fourth string

\ === Heading & Compass Directions
\ 9  Compass Directions & Heading Rotation with Keys
\ turtle cycles through the directions (cw, ccw)  (cw from N to NW rotation)
\ cw: N > NE > E > SE > S > SW > W > NW
\ ccw: opposite
\ There are 8 states rotating -- Heading is an integer 0 - 7 which indexes into a state array and gets the heading character
\ The maths of rotation is mod, so this is arithmetic mod-8

create heading 1 bytes allot  \ 0 to 7
: set-heading ( h# -- ) heading c! ;
: init-heading ( -- ) 0 set-heading ;  \ default heading is N=0
: get-heading ( -- h# ) heading c@ ;
: adjust-heading ( h-delta -- )  get-heading + 8 mod set-heading ;  \ each h-delta nudges cw by 45 degrees, negative ccw by 45 deg

\ graphical symbols to show what direction turtle is heading - numbers are stored little endian (small byte first)...
$5E constant  NN   \    ^  alternates: $70 p  $56 V  $5E ^
$2F constant  NE   \    / 
$3E constant  EE   \    >  
$5C constant  SE   \    \  
$76 constant  SS   \    v  alternates: $62 b  $59 Y  $56 V
$2F constant  SW   \    / 
$3C constant  WW   \    < 
$5C constant  NW   \    \ 
: create-heading-symbol
    create ( <name> -- )
     NN c, NE c, EE c, SE c, SS c, SW c, WW c, NW c,   \ headings in cyclic order 
    does> ( h# -- asc )
        + c@ ;   
create-heading-symbol  heading-symbol

\ direction names, letters backwards because numbers are stored in lower-endian form, i.e. low byte first
$204E  constant  N-name    
$454E  constant  NE-name   
$2045  constant  E-name    
$4553  constant  SE-name   
$2053  constant  S-name    
$5753  constant  SW-name   
$2057  constant  W-name    
$574E  constant  NW-name   
: create-heading-name
    create ( <name> -- )
        N-name , NE-name , E-name , SE-name , S-name , SW-name , W-name , NW-name ,
    does> ( h# -- ptr )
        swap cells + ;
create-heading-name  heading-name

\ =========== Graphics.
\ VGA in DOS is a quick way to produce graphics because in DOS it is easy to access screen pixels (peek/poke) directly via VGA card memory.  
\ This is no \ longer the case in modern windows which requires going through a graphics library such as SDL or OpenGL.
\ This program is written F-PC (16-bit Forth) because F-PC makes it easy to write 8086 assembly code, making VGA on DOS straightforward. 
\ Modern Windows computer?  No problem, use a DOS emulator (vDosPlus recommended or DOSBox).

\ Standard text mode graphics works with 80 columns and 25 rows (2000 cells, indices x:0..79, y:0..24 ).  
\ Implementation of each row is with 160 data columns, so 4000 cells.  
\ Drawing works correctly only on even columns so in practice the 80 column indices are  multiples of 2, i.e. 0,2,4,...,158.
\ Pixel locations are specified in linear order from 0 to 3998d ($F9E in hex) with the final 3999d cell not visible (even pixel only issue).

\ Two representations are possible: modified Cartesian x,y as across and *down* vs. Matrix: i,j for row and column (down/across).
\ THIS PROGRAM uses Cartesian (x,y / across/down).

\ We have THREE coordinate systems: 
\     USER COORDINATES - what the user can access on the screen, i.e. excluding menus, borders, etc., with 0,0 as the origin point for the user's cursor.
\           i.e. within a window, (i.e. a subset of the 80x25 full-screen)
\     SCREEN COORDINATES - where this all happens on the screen, with (0,0) at top left corner of the screen
\     PHYSICAL COORDINATES - the underlying implementation / linear data array against which the drawing pixels must be referenced.
\   xy2lin converts from USER coordinates from physical position in linear array.  lin2xy goes from physical to user.
\   xy2lin-screen converts from SCREEN coordinates to physical.  No need for back-conversion.

\ Example - FPC uses rows 0-1 (1st line is Forth bar, and 2nd line is erased after command scrolls screen up), so to see a pixel within F-PC start at linear position 320 (start of row 3)
\ FPC uses columns 69-79 of the first 12 rows for dictionary, with the full 80 cols starting from row 13.

\   References:
\   Video programming:  http://ece.wpi.edu/~wrm/Courses/EE3803/Labs/roehrl.html
\   Video mode:         http://www.wagemakers.be/english/doc/vga

\ PC Video Mode (BIOS settings)
$1 constant text40x25-8color
$3 constant text80x25-8color            \ text mode (8 color)
$8 constant text80x25-16color           \ default F-PC mode, colsxrows, each character is 8x8 pixels (640 pixels x 200)
$4 constant graphCGA-320x200-8color     \ standard resolution
$10 constant graphEGA-640x350-16color   \ high resolution
$12 constant graphVGA-640x480-16color   \ default VGA graphics mode

code set-vid-mode ( mode -- )   \ calls PC BIOS instruction to set desired video mode
    pop ax          \ AL=video mode 
    mov ah, # 0     \ AH=set video mode (BIOS instruction)
    int $10		    \ DOS interrupt: call BIOS instruction
next end-code

: set-video-mode ( -- ) text80x25-16color set-vid-mode ;
\ : set-video-mode ( -- ) graphVGA-640x480-16color set-vid-mode ;

\ === Physical (linear array) coordinates for text mode: 80x25 = 2000 visible cells.  160x25 = 4000 implementation cells.
  80  constant screen-cols
 160  constant physical-cols  \ note: odd slots are invalid display locations, so counter runs 0,2,4,..,178   (see graphics discussion above)
  25  constant screen-rows
   0  constant physical-zero
3998  constant physical-end    \ $F9E hex
\ From PHYSICAL to SCREEN coordinates
 1  constant top-menu-height
 1  constant bottom-menu-height
 1  constant buffer-thickness  \ frame all round

\ === USER area in SCREEN coordinates (0,0 is top left pixel of the screen)
 0 top-menu-height + buffer-thickness +                  constant screen-top
 screen-rows bottom-menu-height -  buffer-thickness - 1 -   constant screen-bot
 buffer-thickness                                        constant screen-left
 screen-cols buffer-thickness - 1 -                         constant screen-right
screen-bot screen-top - 1 +                              constant screen-height
screen-right screen-left - 1 +                           constant screen-width
 3 1 - physical-cols *                                       constant row#3             \ USER COORDINATE STARTS HERE 3rd line ($140=320d) is one accessible in  user mode (1st is menu, 2nd is border)

\ === Linear offset in physical coordinates to 0,0 in user coordinates (top left pixel that the user can access, i.e. excluding menus and buffers)
 create lin-offset 1 cells allot  \ VARIABLE
: set-lin-offset ( -- ) screen-top physical-cols * screen-left 2 * + lin-offset ! ;
: get-lin-offset ( -- offset ) lin-offset @ ;

: set-interactive-graphics-mode ( -- )  row#3 lin-offset ! ;
set-interactive-graphics-mode       \ default program mode (for testing) 

\ === Coordinate conversion
\ Screen coordinates are Cartesian ( x horizontal, y vertical), with graphics convention (0,0 is top left corner, y is down).  Not using matrix convention ( row i, col j).  
\ Physical Implementation is in linear array, row by row (aka physical coordinates)
\ User coordinates are subset of Screen coordinates, with offsets that prevent user from moving into forbidden areas (menus, etc.)

: xy2lin ( x y -- lin )  \ USER to physical coordinates
\ test: 0 0 xy2lin  = 320, is 3rd row down
    physical-cols *     \ y rows to linear array
    swap 2 *        \ x cols (0..79) to even x-slots (0,2,4,...158) since only evens are visible
    +               \ combine to get linear position
    get-lin-offset + ; \ plus linear offset to reference to user coordinates (0,0) 
: lin2xy ( lin -- x y )  \ physical to USER coordinates
    get-lin-offset -   \ remove offset
    physical-cols /mod ( x y ) 
    swap 2 /          \ to screen char coords (80 per row)
    swap ;
: xy2lin-screen ( x y -- lin )  physical-cols * swap 2 * + ; \ SCREEN to physical coordinates

create cursorpos 1 cells allot   \ VARIABLE - holds physical location of cursor
: setpos ( x y -- ) xy2lin cursorpos ! ;
: getpos ( -- x y ) cursorpos @ lin2xy ;
: init-pen-position ( -- x y ) 0 screen-height 1 - 2dup setpos ;
: print-location ( x y -- x y )   2dup ." (x: " swap . ." y: " . ." )" ;    \ TO DO: Add this to status bar

\ === Painting to Screen- Low-level Character Routines
\ color information is encoded into a single byte for use in assembly language routine.  
\ color-code (e.g. $F4) holds background color (bgcolor) in high nybble ($F is white) and foreground (text) color in low nybble ($4 is red)
\ pixel position is in linear encoded form, i.e. row/col are encoded within a single wrap-around array of screen positions

code poke-pixel ( linear-location  color-code  ascii-code  -- )  \ sets a pixel at physical position in linear array with given colors and ascii character.  
    pop cx                  \ load CL with ascii char
    pop ax                  \ load AL with color-code (foreground & background)
    mov ch, al              \ create text pixel (color|char) in (CH|CL)
    pop bx                  \ load BX with screen location
    push ds                 \ preserve data stack
    mov ax, # $B800         \ load AX with graphics address B800 hex (vga memory base address for direct display to screen)
    mov ds, ax              \ set data segment to graphics address (cannot load literal directly, only via register)
    mov 0 [bx], cx          \ display pixel in desired location in graphics memory [B800:bx] (0 data segment offset)
    pop ds                  \ restore data stack
next end-code
\ test: 1000 $FF BL poke-pixel
\ test: 1000 228 42
\ test: 320 228 42 
\ test: 77 20 xy2lin 228 42

code peek-pixel ( linear-location -- color-code ascii-code )
    pop bx              \ load BL with screen location
    push ds             \ preserve data stack
    mov ax, # $B800     \ load AX with graphics address (B800 hex)
    mov ds, ax          \ set data segment to graphics address
    mov cx, 0 [bx]      \ load CX with pixel data from desired location in graphics memory (B800:bx)  (CH|CL)=(color|char)
    pop ds              \ restore data stack
    mov ax, # 0         \ clear AX
    mov al, ch          \ load AX with color-code
    mov ch, # 0         \ clear CH, leaving CL with ascii code
    push ax             \ return color-code
    push cx             \ return ascii code    
next end-code

: set-char ( linloc asc fgcol bgcol -- ) fgbg2col swap poke-pixel ;    \ test: row#3 '*' red yellow
: poke-xy  ( asc fgol bgcol x y -- ) xy2lin 3 roll 3 roll 3 roll set-char ;   \ FIXES UNFORTUNATE STACK DECISION where location is before character and color
: poke-xy-screen  ( asc fgol bgcol x y -- ) xy2lin-screen 3 roll 3 roll 3 roll set-char ;  \ screen to pixel

: get-char ( linloc -- asc fgcol bgcol ) peek-pixel swap col2fgbg ;    \ test: turtle-init 0 20 read-screen-pixel
: read-screen-pixel ( x y -- asc fgcolor bgcolor ) xy2lin get-char ;

: fill-screen ( color start-loc end-loc -- )  \ color (0-15) or color-name ( black, white, yellow, red, ... )
    rot dup fgbg2col -rot                     \ create color code 
    BL -rot                                   \ fill with blanks    
    over - 2/ 1 +                             \ length of physical array to fill but only the evens (odd cells go green in this video mode) 
( colorfgbg BL start-loc length )             
    0 do ( colorfgbg BL start-loc )           \ Only start-loc needs to change, rest are constant.  Looped instructions: 7.
        >r 2dup r@ -rot poke-pixel r> 2 +     \ increment by 2 as only even cells are visible
    loop 3drop ;  

: black-screen  ( -- ) black physical-zero physical-end fill-screen ;

create border-color 1 bytes allot 
: set-border-color ( color -- ) border-color c! ;
: get-border-color ( -- color ) border-color c@ ;
silver set-border-color 

: hline-draw ( color x1 x2 y -- )
    -rot swap do ( color y ) 
        i over xy2lin-screen bl 3 pick dup set-char
    loop
    2drop ;
: vline-draw ( color x y1 y2 -- )
    swap do ( color x )
        dup i xy2lin-screen bl 3 pick dup set-char
    loop
    2drop ;
: borders-draw ( color -- )  \ all coords are in physical space
    get-border-color 0 80 1 hline-draw
    get-border-color 0 80 23 hline-draw
    get-border-color 0 1 23 vline-draw
    get-border-color 79 1 23 vline-draw ;


\ === PIXELS & PICTURE
\ A pixel is a 3 byte triple of ascii code, foreground color, background color, stored in that order.
\ Picture is a single array (nrow x ncol x 3 bytes) of pixels used to save/restore picture to/from memory.
\ Pixel variable is a temporary storage to hold 1 pixel information while user is deciding, in case the picture gets impacted (e.g. home, penup/down, etc.)
3   constant   pixel-sz
create         pixel          pixel-sz   bytes allot                       \ variable - temp storage for 1 pixel
screen-cols screen-rows *     constant   num-pix-scr
num-pix-scr    pixel-sz *     constant   pic-sz               
create                                   picture      pic-sz bytes allot    \ array - memory block to hold full screen

: picture-clear ( --- )  picture pic-sz 0 fill ;   \ zero out memory
picture-clear

: pix!+ ( asc fgcolor bgcolor pix-ptr -- pix-ptr' ) >r rot r> c!+ rot swap c!+ c!+ ;    \ store preserving pixel order
: pix@+ ( pix-ptr -- asc fgcolor bgcolor pixl-ptr' ) c@+ swap c@+ swap c@+ swap ;
: cache-pixel ( x y -- ) read-screen-pixel  pixel  pix!+  drop  ;

\ Changing a pixel
\ When turtle is on the pixel (show-heading), where turtle has moved on from with pendown (paint-xy) or with penup (restore-xy), where backspace (erase-xy) 
: show-heading  ( x y h# -- )   -rot xy2lin swap heading-symbol get-pen-color get-cursor-color set-char ;  \ draw symbol on screen 
: paint-xy      ( x y -- x y )  2dup xy2lin BL get-pen-color dup ( loc asc fg bg ) set-char ; \ test 3 12
: restore-xy    ( x y -- x y )  2dup xy2lin pixel pix@+ drop set-char ;
: erase-xy      ( x y -- x y )  2dup xy2lin BL white white set-char ;

\ Changing a picture: note each pixel is a byte in memory, but occupies two bytes on screen.
: pic-scr2mem ( -- ) picture num-pix-scr 0 do i 2 * ( pic linloc ) swap >r get-char r> pix!+ loop drop ;   \ save picture from screen to memory
: pic-mem2scr ( -- ) picture num-pix-scr 0 do i 2 * swap pix@+ >R set-char R> loop drop ;     \ load memory from memory to screen

\ === Paint to Menu
\ !! HOW IT SHOULD WORK (how it works seems more complicated)
\ The menu display system should be simple: draw the menu, refresh dynamic content when it changes.
\ Do this to a virtual copy of the menu so that it persists even with screen changes (e.g. display help system, clear screen, etc.)
\ Text should use the default fgbg unless in a special mode, in which case this is provided instead.
\ !! END CODE REVIEW

\ Menu Features - screen coordinates
0   constant menu2-h#      \ heading (x=.)
13  constant menu2-c#      \ color (x=.)
24  constant menu2-r#      \ recording (x=.)
0   constant menu1-row     \ top menu (y=.)
24  constant menu2-row     \ bottom menu (y=.)

: menu1-text ( -- )  " Commands: [ARROWS]Move/Rotate [BKSPC]Erase [']Penup/Dn [H]ome [Esc]Clear [?]HELP" ;  \ 80 char max
: menu2-text ( -- )  " -- [X/C]olor:------ [T]:------------ [N]extSkill [1-9]ShowSkill           [Q]uit" ;  \ 80 char max   \ [L]oop:--

\ TESTER CODE FOR BETTER WRITING OF MENUS
( linloc c-adr nchar fg bg -- )

\ writing characters and strings
: write-char+ ( col L ca  -- col L' ca' ) 
        over 2 + -rot ( col l' l c )            \ save next location (2 bytes ahead - see graphics explanation: 80 screen positions, 160 physical positions)
        dup 1 + -rot c@ ( col l' c' l char )    \ get char and save next pointer position
        4 pick swap ( col l' c' l col char )    \ arrange parameter order
        poke-pixel ;
: write-string ( L ca n fg bg -- ) fgbg2col swap >r -rot r> ( col L ca n ) 0 do write-char+ loop ;
: test-setup 0 15 xy2lin-screen menu1-text red lime ;  \ the 0 15 are just the menu row and col (screen coords) for starting the string write
: test test-setup write-string ;
\ CONSIDER changing the parameter order to set-char

\ menu parameters: row location (y), foreground/background color for text, and col position on menu to update dynamic information (in screen coordinates)
: menu1-row 0 ;
: menu1-fgbg  white navy ;
: menu2-row 24 ;
: menu2-fgbg  white navy ;
: menu2-heading-position ( -- loc )  menu2-h# menu2-row  ;
: menu2-color-position ( -- loc )    menu2-c# menu2-row  ;  
: menu2-rec-pos ( -- loc) menu2-r# menu2-row ;
: create-rec-fgbg      \ based on rec state
    create 
        white c, slate c,     \ if 0 
        red c, yellow c,      \ if 1
        navy c, yellow c,     \ if 2
        white c, red c,       \ if 3
    does> ( -- )
       get-rec-state 2 * bytes + c@+ swap c@ ; 
create-rec-fgbg menu2-rec-fgbg
: menu2-attn-fgbg ( -- ) red lime ;  \ color scheme to highlight text

\ creating menus
: new-menu 
    create ( menu-text_c-addr|nchar  menu-row  fg bg  | <name> -- )   \ create dictionary entry
    fgbg2col c,                                                       \ create default menu color code
    physical-cols * ,                                                 \ physical offset to start on screen at (0,menu-row)
    here dup , screen-cols 2dup dup bytes allot BL fill drop swap cmove   \ allocate memory for string, initialized with blanks, then copied with menu-text
does>  ( -- )                                                          \  places address of entry on stack 
    c@+ swap  \ fgbg
    @+ swap   \ start-loc physical   
( col L ca )
    screen-cols 0 do write-char+ loop                                  \ print menu to screen
    3drop ;

menu1-text  menu1-row  menu1-fgbg   new-menu  menu1
menu2-text  menu2-row  menu2-fgbg   new-menu  menu2

: test-pixel  1000 31 'X' poke-pixel ;
: setup-menutest menu2-text menu2-row menu2-fgbg ;
: test-create ( <menu3> ) setup-menutest new-menu ;

\ dynamically updating menus
: menu-write ( c-addr length fg bg xcol-screen yrow-screen -- )
    xy2lin-screen             \ calc physical offset
    -rot >r >r -rot r> r>     \ prepare stack
    ( linloc c-addr nchar fg bg )
    write-string 3drop ;
: test-menu-write  5 color-names red yellow 20 23 menu-write ;

: menu2-show-heading ( -- )  get-heading heading-name  2  menu2-fgbg menu2-h# menu2-row menu-write ;
: get-pencolor-name ( -- index ) get-penup-state if penup-col-index else get-pen-color then color-names ;
: menu2-show-color ( -- )  get-pencolor-name get-penup-state if menu2-attn-fgbg else menu2-fgbg then menu2-c# menu2-row menu-write ;
: menu2-show-recording-state ( -- )  get-rec-str  menu2-rec-fgbg  menu2-r# menu2-row menu-write ;

: menu2-refresh ( -- ) menu2-show-heading  menu2-show-color  menu2-show-recording-state ;  \ paints the dynamic portions
: paint-menus  ( -- ) menu1 menu2 ;  \ paints static menus
: paint-screen ( color -- ) physical-zero physical-end fill-screen paint-menus borders-draw ;

\ TESTER CODE
: menu2-screen-pos ( x y -- ... ) ;

\ === Show/Hide Turtle
create visible-state 1 cells allot  \ VARIABLE - BOOLEAN - turtle visibility state -- true if visible, false if hidden
: set-visible ( -- ) true visible-state ! ;
: set-hidden  ( -- ) false visible-state ! ;
: is-visible? ( -- t/f) visible-state @ ;
: turtle-hide ( x y -- x y )  restore-xy ;   \ erase current location
: turtle-show ( x y -- x y )  is-visible? if 2dup get-heading show-heading else turtle-hide then ;

\ === Movement Control
\ movement is either forward or backwards along the axis of the current heading, with screen wrap-around
\ moving forward paints, moving backward erases
\ coordinates (x,y) are semi-cartesian: over and down.  (0,0) is the top left of the user space, i.e. excluding the screen border and menus.
\ movement = 1) update coordinates to a new pixel, 2) paint or restore previous pixel, 3) new pixel is cached, 4) show turtle in new pixel.

\ screen wrap-around
: wrap-y ( y -- y* ) screen-height mod ;
: wrap-x ( x -- x* ) screen-width mod ;

\ movement along direction axes
: move-N-fwd   ( x y -- x' y' ) 1 - wrap-y ;
: move-N-back  ( x y -- x' y' ) 1 + wrap-y ;
: move-E-fwd   ( x y -- x' y' ) swap 1 + wrap-x swap ;
: move-E-back  ( x y -- x' y' ) swap 1 - wrap-x swap ;
: move-S-fwd   ( x y -- x' y' ) move-N-back ;
: move-S-back  ( x y -- x' y' ) move-N-fwd ;
: move-W-fwd   ( x y -- x' y' ) move-E-back ;
: move-W-back  ( x y -- x' y' ) move-E-fwd ;
: move-NE-fwd  ( x y -- x' y' ) move-N-fwd move-E-fwd ;
: move-NE-back ( x y -- x' y' ) move-N-back move-E-back ;
: move-SW-fwd  ( x y -- x' y' ) move-NE-back ;
: move-SW-back ( x y -- x' y' ) move-NE-fwd ;
: move-NW-fwd  ( x y -- x' y' ) move-N-fwd move-W-fwd ;
: move-NW-back ( x y -- x' y' ) move-N-back move-W-back ;
: move-SE-fwd  ( x y -- x' y' ) move-NW-back ;
: move-SE-back ( x y -- x' y' ) move-NW-fwd ;

\ stores xt for array of movements in h-index (rotation cycle) order
: create-movement-vector  
    create ( nw w sw s se e ne n <name> -- )
        , , , , , , , ,  \ store 8 xts
    does> ( x y -- x' y' )
        get-heading cells + @ execute ;

' move-NW-fwd   ' move-W-fwd    ' move-SW-fwd    ' move-S-fwd    
' move-SE-fwd    ' move-E-fwd    ' move-NE-fwd    ' move-N-fwd   
create-movement-vector   move-fwd-headings       

' move-NW-back    ' move-W-back    ' move-SW-back    ' move-S-back    
' move-SE-back    ' move-E-back    ' move-NE-back    ' move-N-back
create-movement-vector   move-back-headings

: move-fwd ( x y -- x' y' )  get-penup-state if restore-xy else paint-xy then  move-fwd-headings  ;
: move-back ( x y -- x' y' )  erase-xy  move-back-headings  ;  \ erases whether in penup mode or not
                                 
\ === Key logger / Diary function
9 constant num-diaries                  \ number of distinct recordable macros
create active-diary-num 1 bytes allot   \ digits 0 to 8
: set-active-diary-num ( digit -- ) dup active-diary-num c! update-diary-string ;
: get-active-diary-num ( -- digit ) active-diary-num c@ ;
0 set-active-diary-num  \ default diary = 0
: next-diary ( -- ) 
    get-active-diary-num 1 + num-diaries mod  ( digit )
    set-active-diary-num  ;

create preserved-diary-num 1 bytes allot  \ used to hold the original diary number so that nested macro playbacks end by showing the original diary number.
: store-active-diary-num ( -- ) get-active-diary-num preserved-diary-num c! ;
: restore-active-diary-num ( -- ) preserved-diary-num c@ set-active-diary-num ;

1000 bytes constant diary-length    \ max number of key presses (bytes) that can be recorded in a single macro
num-diaries diary-length * constant diary-length-total
create   diary   diary-length-total  allot      \ one array containing all diaries
diary diary-length-total 0 fill   \ zero out mem 

create diary-pos num-diaries cells allot  \ array holding next free place in each diary, which is the stop point for playback .
diary-pos num-diaries cells 0 fill \ zero out mem
: get-diary-pos (  -- dentry ) get-active-diary-num cells diary-pos + @ ;

create diary-end num-diaries cells allot                       \ array holding diary ends 
diary-end num-diaries cells 0 fill \ zero out mem
: get-diary-end ( -- dentry ) get-active-diary-num cells diary-end + @ ;
: diary-full?  ( dpos -- t/f ) get-diary-end > ;

: set-diary-end ( -- )  \ store diary end position into each cell 
    diary 1 - 
    num-diaries 0 do 
        diary-length + dup 
        i cells diary-end + ! 
    loop drop ; \ arithmetic increments since diaries are consecutive in memory
: set-diary-pos ( dentry -- )  get-active-diary-num cells diary-pos + ! ;
: cue-diary-start ( -- diary-ptr ) get-active-diary-num bytes diary-length * diary + ;  \ returns start position
: init-diary ( -- ) cue-diary-start set-diary-pos ;
: init-diaries ( -- ) 9 0 do init-diary next-diary loop ;
: rewind-diary-pos ( -- ) get-diary-pos 1 - set-diary-pos ;  \ removes last entry

create diary-recording 1 cells allot   \ state variable - true if recording
: is-recording? ( -- t/f ) diary-recording @ ;
: set-diary-recording ( -- ) true diary-recording ! ;
: clear-diary-recording ( -- ) false diary-recording ! ;
: show-recording ( -- )  rec-state-recording  set-rec-state  menu2-show-recording-state ;
: flash-recording ( -- )  rec-state-reminder  set-rec-state  menu2-show-recording-state 300 ms show-recording ;

: start-recording ( -- ) set-diary-recording       show-recording                    init-diary ;  \ record macro
: stop-recording ( -- )  clear-diary-recording     rec-state-ready  set-rec-state    menu2-show-recording-state ;
: warn-diary-full ( -- ) 
    rec-state-full set-rec-state menu2-show-recording-state
    700 ms stop-recording ;
: diary-write-entry ( asc -- )  get-diary-pos c!+ dup diary-full? if drop warn-diary-full else set-diary-pos then ;

\ === Key Press & Macro processing

\ We find what action turtle should take based on user keypress using a key mapping lookup table which associates every ascii key with the executable token (xt) giving its intended action (if it has one).
\ The actual lookup table is below, after the actions have all been defined (search =keymap=)
create   key-map   256 cells allot              \ key mappings are (key xt)
: add-key ( xt key -- ) cells key-map + ! ;     \ store executable token (xt)in key-map table, indexed by ascii keypress code
\ Non mapped keys are ignored.
: ignore-input  ( asc -- ) drop ;
' ignore-input  constant xt-ignore-input 
: init-key-map ( -- ) 255 0 do xt-ignore-input i add-key loop ;  \ default is to ignore input; 

\ This is an advanced Forth construction (create-does).  It compiles a data structure within an instruction for the object to use (data+action)
: create-mapping  ( key-map <name> -- )
    create ,                \ compiles key-map base address into <named> instruction
  does> ( x y asc -- x y )  \ defines what the created instruction (e.g. turtle-do) will do when called
    @ ( x y asc key-map ) over cells + @ ( x y asc xt ) execute ;
key-map create-mapping turtle-do

\ This is the key processing engine
: key-process ( x y asc -- x y ) turtle-do turtle-show menu2-refresh ;

: which-diary? ( asc - d# ) asc2digit 1 - ;
: macro-playback ( x y asc -- x' y' )
    store-active-diary-num 
    which-diary? set-active-diary-num
    cue-diary-start
    begin ( x y diary-pos )   \ read from start position up to current position
        dup get-diary-pos < 
\ pic-scr2mem  \ scaffolding
    while
        c@+ ( ptr asc ) swap >R 
\            cr dup show-key \ scaffolding
            key-process 
            R>  \ restore next pointer
    repeat
    \ pic-mem2scr  \ scaffolding
    drop
    restore-active-diary-num ;

\ === TURTLE PROCESSING COMMANDS
\ These commands are ALWAYS logged if recording is active.
: log-key-if-recording ( asc -- asc ) is-recording? if dup diary-write-entry then ;
: turtle-move ( x y asc -- x' y' )
    log-key-if-recording ( asc -- asc )
    fwd-key = if move-fwd else move-back then
    2dup cache-pixel 2dup setpos
    set-visible ;
: heading-change ( asc -- )
    log-key-if-recording ( asc -- )
    cw-key = if 1 else -1 then adjust-heading
    set-visible ;
: penupdown ( x y asc -- x y ) log-key-if-recording  drop get-penup-state if do-pendown else do-penup then ;
: color-change ( x y asc -- x y ) log-key-if-recording   get-penup-state if do-pendown drop else turtle-color then ;


\ These commands are SOMETIMES logged if recording is active.
\ One level nesting of recordings is allowed, but not higher level and not recursion.   LIMITATION (BUG) - only 1 level nesting is allowed??
\ Nesting: i.e. play-macro instruction is logged if recording without playing; else play.  
\ Recursion (ont allowed): Won't log instruction to play macro when record-to diary is same current active diary
\ First class nesting --- combining instructions and macros containing only instructions (first class macros) - ALLOWED
\ Higher level nesting --- combining macros that themselves contain macros.
: log-key-if-not-current-diary ( asc -- asc ) dup which-diary? get-active-diary-num = if flash-recording else dup diary-write-entry then ;
: play-macro ( x y asc -- x y )  is-recording?  if log-key-if-not-current-diary drop else macro-playback then ; 


\ These commands are not logged.
: turtle-showhide ( x y asc -- x y ) drop is-recording? 
    if flash-recording 
    else is-visible? if set-hidden else set-visible then 
    then ; \ toggle turtle visibility
: turtle-home ( -- )
    get-penup-state dup >r                         \ store penup-state
    if else do-penup  then turtle-hide             \ don't leave mark in last position 
    2drop init-pen-position init-heading 
    2dup cache-pixel                               \ don't clobber home position
    r> if else do-pendown then ;                   \ usability: restore the pen to the user's last state
: go-home ( x y asc -- x y ) drop is-recording? if flash-recording else turtle-home then ;    \ reset turtle to home position

: diary-next ( x y asc -- x y ) drop is-recording? if flash-recording else next-diary then ;
: recording-onoff ( x y asc -- x y )  drop is-recording? if stop-recording else start-recording then ;

: safe-cr ( -- ) pic-scr2mem cr pic-mem2scr ;     \ buffering and redrawing screen to avoid damage from carriage return
: confirm-quit ( -- t/f )
    safe-cr  prompt-quit  quit? if true else cr pic-mem2scr false then ;
: is-quit? ( asc -- asc t/f ) 
        quit-key over <> over quit-key2 <> and  ( asc t/f )  \ checking both cases (q and Q)
        if false                                             \ no quit request, continue
        else                                                 \ quit request
           is-recording? 
                if flash-recording false                     \ not allowed to quit while recording
                else confirm-quit if true else false then
                then
        then ;
: clear-screen ( x y asc -- x y ) drop  is-recording? 
    if flash-recording 
    else
        safe-cr                        \ save screen in case user changes mind
        clear? cr if  
            get-screen-color  cr paint-screen turtle-home set-visible 
        else  cr pic-mem2scr                                                \ restore saved screen
        then  
    then ;
        
: show-help ( -- )  pic-scr2mem  black-screen  welcome pause-key  pic-mem2scr ;
: help-screen ( x y asc -- x y ) drop is-recording? if flash-recording else show-help then ;

\ === Initialization code
: turtle-init ( -- x y )
    set-video-mode
    set-lin-offset
    set-fgbg-palette   \ assign optimized cursor/background colors
    0 set-active-diary-num   init-diaries  set-diary-end  
    attach-rec-msgs   stop-recording
    init-heading   init-pen-position ( -- x y ) 
    clear-penup-state          \ default = pen-down
    red set-pen-color          \ default pen & bg color
    white set-screen-color     \ default screen color 
    silver set-border-color 
    cr
    white paint-screen menu2-refresh
    2dup cache-pixel            \ hold the original color
    set-visible 
    turtle-show ;

\ =keymap=
init-key-map
\ Mapped actions overwrites the ignore
\         xt               key
    ' turtle-move       fwd-key              add-key
    ' turtle-move       erase-key            add-key
    ' heading-change    left-key             add-key
    ' heading-change    right-key            add-key
    ' penupdown         penupdown-key        add-key
    ' color-change      color-fwd-key        add-key
    ' color-change      color-fwd-key2       add-key
    ' color-change      color-back-key       add-key
    ' color-change      color-back-key2      add-key
    ' turtle-showhide     show-hide-key        add-key
    ' go-home           home-key             add-key
    ' go-home           home-key2            add-key
    ' diary-next      diary-key            add-key
    ' diary-next      diary-key2           add-key
    ' recording-onoff  record-key           add-key
    ' recording-onoff  record-key2          add-key
    ' play-macro        play-macro-1-key     add-key
    ' play-macro        play-2-key           add-key
    ' play-macro        play-3-key           add-key
    ' play-macro        play-4-key           add-key
    ' play-macro        play-5-key           add-key
    ' play-macro        play-6-key           add-key
    ' play-macro        play-7-key           add-key
    ' play-macro        play-8-key           add-key
    ' play-macro        play-9-key           add-key
    ' clear-screen      clear-screen-key     add-key
    ' help-screen       help-key             add-key
    ' help-screen       help-key2            add-key


\ === Main program
: q ( -- ) \ quit
    is-recording? if stop-recording then
        cr ." To restart, type TURTLE-LOGO"
        cr ." To return to Windows, type EXIT." 
\        bye  \ exits Forth to DOS.  Comment out to stay within Forth e.g. during development
        ;  

: turtle-go ( -- x y )  
    show-help  
    turtle-init
    begin ( x y )              \ user input processing loop
        key ( x y asc )
        is-quit? if drop true else key-process false then  
    until ( x y )
    turtle-hide  is-recording? if stop-recording then q ;
: turtle-logo ( -- ) turtle-go ;   \ alternative

\ ================================================ END OF MAIN PROGRAM



\ ================================================ TESTING - SUPPLEMENTARY CODES
\ === instructions
: kp ( x y asc -- x y ) key-process  ; 
: .. ( -- ) pause-quit ;
: pen-adjust ( -- ) penupdown-key kp ;
: recording-adjust ( -- ) record-key kp ;
: diary-adjust ( -- ) diary-key kp ;
: turtle-adjust ( -- ) show-hide-key kp ;
: test-home ( -- ) home-key kp ;
: next-color ( -- ) color-fwd-key kp ;
: prev-color ( -- ) color-back-key kp ;

\ === macros
: left-45 ( -- ) left-key kp ;
: left-90 ( -- ) 2 0 do left-45 loop ;
: left-180 ( -- ) 4 0 do left-45 loop ;
: right-45 ( -- ) right-key kp ;
: right-90 ( -- ) 2 0 do right-45 loop ;
: right-180 ( -- ) 4 0 do right-45 loop ;
: fwd-n ( n -- ) 0 do fwd-key kp loop ;
: erase-n ( n -- ) 0 do erase-key kp loop ;

: draw-half-square ( -- )  2 fwd-n   right-90  5 fwd-n  right-90 ;
: house-square ( -- ) draw-half-square draw-half-square ;
: house-roof ( -- ) 3 fwd-n right-45 2 fwd-n right-45 1 fwd-n right-45 2 fwd-n right-45 1 fwd-n ;
: house-complete ( -- ) pen-adjust 2 fwd-n left-90 3 fwd-n left-90 pen-adjust ;
: draw-house ( -- ) house-square house-roof house-complete ;
: draw-spiral ( -- ) 
    4 fwd-n   right-90  8 fwd-n  right-90
    3 fwd-n   right-90  1 fwd-n  color-fwd-key kp  5 fwd-n  right-90 ;

\ === test harnesses
: test-show-color-names ( -- ) 17 0 do i color-names cr type loop ;

: test-poke-pixel ( -- )
    row#3 $5F 'S' poke-pixel  \ interactive visible start point
    physical-end $8F 'E' poke-pixel      \ end
    3 1 xy2lin $A0 'A' poke-pixel   
    68 12 xy2lin $C0 'B' poke-pixel
    69 13 xy2lin $40 'C' poke-pixel ;

: test-probe-diary  ( -- )
    0 set-active-diary-num
    init-diaries  set-diary-end
    ." diary # " get-active-diary-num . 
    ." diary-pos: "  get-diary-pos u. 
    ." diary-end: "  get-diary-end u. 
    cr ;

: test-diaries ( -- )  
    1 fwd-n '1' kp ..   \ checking that no playbacks occur, no errors occur, but diary changes with each keypress
    1 fwd-n '2' kp ..
    1 fwd-n '3' kp ..
    1 fwd-n '4' kp ..
    1 fwd-n '5' kp ..
    1 fwd-n '6' kp ..
    1 fwd-n '7' kp ..
    1 fwd-n '8' kp ..
    1 fwd-n '9' kp ..
    right-90
    1 fwd-n '1' kp ..    ;
    
\ TO TEST fullness for diaries, set diaries from 1000 to 15.  Remember to set the diary length back to 1000 chars.
: test-diary-fullness-1-channel ( -- )  \ check that all 9 diaries work correctly.
    record-key kp ..
    8 fwd-n right-90 ..  \ 10 moves
    1 fwd-n right-90 next-color ..  \ 14 moves 
    right-45 .. \ 15th move triggers fullness, does not record, but executes
    left-45 ..  ; \ corrects  
    
: test-diary-fullness ( -- )  \ recording, fullness, and playback.
    turtle-init
    test-diaries
    test-diary-fullness-1-channel diary-key kp ..
    test-diary-fullness-1-channel diary-key kp ..
    test-diary-fullness-1-channel diary-key kp ..
    test-diary-fullness-1-channel diary-key kp ..
    test-diary-fullness-1-channel diary-key kp ..
    test-diary-fullness-1-channel diary-key kp ..
    test-diary-fullness-1-channel diary-key kp ..
    test-diary-fullness-1-channel diary-key kp ..
    test-diary-fullness-1-channel diary-key kp ..
    '1' kp ..
    '2' kp ..
    '3' kp ..
    '4' kp ..
    '5' kp ..
    '6' kp ..
    '7' kp ..
    '8' kp ..
    '9' kp .. ;    

: test-multi-diary-recording ( -- )  \ draws a house all together, then in recordable bits, then by playing the recorded bits, then stitching the recorded bits into one recording
    turtle-init
    draw-house ..
    recording-adjust house-square .. recording-adjust .. diary-adjust ..
    recording-adjust house-roof .. recording-adjust .. diary-adjust ..
    recording-adjust house-complete .. recording-adjust .. diary-adjust ..
\    '1' kp ..
\    '2' kp ..
\    '3' kp ..
    recording-adjust '1' kp .. '2' kp .. '3' kp .. recording-adjust .. diary-adjust ..
    '4' kp .. ;    

    
: test-erase ( -- )
    turtle-init
    right-45    3 fwd-n   3 0 do next-color loop   3 fwd-n   3 0 do next-color loop   3 fwd-n  ..
    right-180 ..
    pen-adjust ..
    7 fwd-n ..
    5 erase-n ..  \ bug!
    right-90 .. 
    1 erase-n .. 
    right-90 .. 
    1 fwd-n .. ; \ bug!

: test-home-key ( -- )
    turtle-init
    draw-spiral draw-spiral draw-spiral ..
    test-home ..
    6 fwd-n pen-adjust ..
    draw-spiral ..
    test-home ..
    right-45 6 fwd-n ..
    test-home ..
    right-45 8 fwd-n ..
    pen-adjust ..
    test-home ..
    right-45 14 fwd-n ..
    test-home ..
    right-45 14 fwd-n pen-adjust ..
    test-home .. ;

: turtle-test ( -- x y ) 
    turtle-init ( -- x y )
\ testing fwd,right,left
    fwd-key kp right-45  2 fwd-n left-90 2 fwd-n ..   
\ testing erase
    right-45 2 fwd-n .. 
    2 erase-n ..  
\ testing colorfwd/colorback
    3 0 do next-color loop 2 fwd-n ..
    2 0 do prev-color loop 2 fwd-n ..
\ testing penup/pendown
    right-90 8 fwd-n ..
    left-180 pen-adjust .. 
    4 fwd-n ..
    pen-adjust ..
    2 0 do next-color loop ..
    2 fwd-n ..
\ testing home
    test-home ..
    left-90 4 fwd-n ..
\ testing macro-key before any record
    5 0 do next-color loop .. 
    play-macro-1-key kp ..  \ nothing bad should happen
    next-color ..
\ testing recording and pressing macro-key while recording
    recording-adjust ..
    play-macro-1-key kp ..  \ recording notification should flash
    next-color .. 
    1 fwd-n .. 1 fwd-n .. 1 fwd-n .. right-90 .. \ one leg of square
    recording-adjust ..
\ testing macro replay
    play-macro-1-key kp ..
    play-macro-1-key kp ..
    play-macro-1-key kp ..
\ testing clear-screen
    clear-screen-key kp  .. \ press yes
    left-180
    play-macro-1-key kp ..
    play-macro-1-key kp ..
    play-macro-1-key kp ..    
    clear-screen-key kp  .. \ press no
\ testing help
    left-90 left-key kp
    play-macro-1-key kp ..
    play-macro-1-key kp ..
    play-macro-1-key kp ..
    play-macro-1-key kp ..
    help-key kp ..
\ testing hide and the various keys that force visibility
    left-90 7 fwd-n ..
    turtle-adjust ..
    next-color ..
    turtle-adjust ..
    1 fwd-n ..
    turtle-adjust ..    
    right-45 .. 
    turtle-adjust .. 
    test-home ..
    turtle-adjust ..
\ test quit
    quit-key kp ;

: fill-screen-test ( -- )
  16 0 do
    i physical-zero physical-end fill-screen pause-quit
  loop ;
4 constant thickness
: color-bg-loc ( lin -- color )  \ pick background color based on location
    lin2xy drop ( j )
    thickness / 16 /mod drop ( color )  ; \ cycle through 16 colors
: color-demo ( -- )  
    set-video-mode
    get-lin-offset
    begin ( loc )
        dup physical-end < 
    while ( loc )
        dup BL over color-bg-loc dup set-char
        2 +  \ only even cells are visible in this graphics mode.  Odd cells go green
    repeat
    drop
    pause-quit ;

: fgbg-demo ( -- )
    set-video-mode
    set-fgbg-palette
    get-lin-offset
    begin ( loc )
        dup physical-end <
    while ( loc)
        dup 0 heading-symbol over color-bg-loc dup get-optim-bg-color set-char
        2 + 
    repeat
    drop
    pause-quit ;

\ =================================================================
\ TODO:
\  2018-08-20--Mon--23:56 - Restarted

\ Usability - Make the SKILL number highlighted and long with the [N] in new skill  OR have it [S]kill #X
\ Feature: Keyboard shortcut list (K mode) -- pulls out a side panel that then goes away.
\ Bug: how to prevent accidentally pressing record and overwriting previously recorded macro?  Prompt with "Are you sure?"  --- That skill has already been used, are you sure?  And pull out the skill code...
\ Bug: Add second class macros (macros that call themselves) -- why does this not work?
    \ BUG: Only 1 level recording nesting is allowed?  So can't do (1,1), (2,2,2) and combine into (1,1,2,2,2)? (Yahtzee feature)
    \ Feature: Allow higher level nesting in which macro calls other macro (but ok still not to have recursion, i.e. macro calling itself which is expressly forbidden)
\ DONE: allow nested recording of macros (i.e. record macro numbers so macros can be nested within other macros and all played back -- BUGGY - need to maintain a recursion stack for reading the macros back )
\ FIXED: cannot do recursion (i.e. cannot record macro calling itself)
    \ Feature: Consider: support recursive macros (calling themselves?)
\ Feature: Add L for loop -- and you enter how many times to issue that instruction.  (L mode).  Can do it stepwise (with press any key).  This is the start of robotic navigation programming.
\ Feature: Add coordinates to status bar (useful for geometry).  Turtle is at 0,0, and he moves into the 1st quadrant.  (G mode) (Geometry mode)
    \ Feature: display numbered grid (numbered rows and columns according to elementary math convention: H is the origin (0,0) then positive to right and up)
    \ Feature: show coordinate location (trouble shoot with .s on status bar)
\ Feature: 'x' to toggle to text mode, to allow adding text in direction of cursor.  Possibly an auto-date feature when saving the picture.
\ Feature: Have turtle prompt children for certain things.
\ Feature: built in exercises -- shows screen, then asks child to replicate it...
\ Feature: record all key presses, and auto-save, so can quit, restore, save, undo full stack, re-do, and reproduce anything
    \ Feature: save macros to file for reloading later (in case program quits).  Allows user to load previous creations, e.g. trees, flowers, birds, sun, house.  
\ Feature: Save picture, screenshot, or generate print-out
    \ Feature: save screen to disk, load from disk
    \ Feature: allow full-screen view - without menus (just the turtle) - or just taking the picture without the menus.
    \ Screenshot: to png
\ Feature: Save and Load files
    \ Feature: Add save and restore keys -- maybe make automatic save just before playing a macro.  So can undo if press the wrong one.
    \ Feature: save and then restore, i.e. in case of a catastrophic error!  write to file.
    \ Also allows creations to be shared and a community of users to be formed.  
    \ Status bar prompts for save to file, load from file, etc.
\ Feature: Undo one instruction
\ Feature: Allow putting letters on the screen
\ Feature: View & debug macros.  Review macro recording instructions, and be able to edit... (programming and debugging instructions!)  (M mode)
\ Installer: Can DOS Box start up with the splash screen?  vDOSbox can play 32 bit files, so using IrfanView in slideshow mode as a simple screenshot.  Later - do this in Turtle Logo itself... or VGA mode, then switch back to text mode.
\ Feature: read in a saved Turtle png and essentially digitize that using turtle (as a way to load files) - basically rasterize turtle's movements using penup, pendown, and color search! :)  image processing...
\ Usability: Think through how colors should work with going forward, backspace, penup, i.e. what is painting, which square, etc.  
\ Usability: Think through which pixel should get colored, etc.  Issue is 2xF creates a square that is 3 pixels high.

\ Feature: design a random maze generator

\ Feature: Port to the Raspberry Pi, Arduino, or 8052.
    \ How big would this program be in assembly language removing all the graphics bits but just retaining the "do turtle" bits with stepper motor control?
    \ Port to Android! (keys will make it harder)  Maybe bluetooth keyboard.  Other zigbee connection with computer so one is controlling remotely with computer keyboard.
    \ Stepper motors
    \ Sensors (bumping into)
\ Feature: Port to 640x380 graphics mode screen instead of text mode (see fire.asm)
\ Feature: Re-mapping the keys to use mnemonics from a non-English language, or extend to create a multi-lingual learning environment
\ PARTIAL DONE through vDOSBox  - Feature: Port to windows via Win32 Forth
\ Feature: voice control of the turtle
\ Feature: sound feedback.  (chiptune if level reached successfully?)
\ Feature: speech synthesizer?

\ Feature: Add ability to animate, e.g. the splash screen images
\ Installer: Tell the user where the installation is going
\ Installer: test the desktop shortcut placement, and the start-menu placement.
\ Application: barack obama ascii art portrait.

\ TODO 2018-08-24--Fri--16:39 --- simplify the menu writing!!!  it is too complicated as it stands.  Use the tester code write-string
\ In progress - buggy - too many items on the stack.

\ BUG --- diary 9 is always full...
\ DONE - !! CODE REVIEW: writing text to screen or menu should be simple and uniform provide text ( ptr nchar), location, and fgbg color -- then write to menu !!
\ FEATURE --- [N/P] next / previous skill.
\ Notify if feature is full.
\ Log *all* key presses, not just when recording, so that can undo/redo/save file/load file.
\ DONE - !! CODE REVIEW: simplified string writing with uniform words for all fields. 
\ TO DO -- consolidate the fixed width string structures into a single data structure...  should it take strings from the input stream and build the structure dynamically?
\ use WORD to simplify the other string structures AND create> does...
\ USABILITY --- let the child doing recording press ESC and it terminates recording safely.  Similarly for other current forbidden actions -- i.e. don't force the child to worry about parentheses at this age :)  it's just frustrating.




